// ---------------------------------------------------@RisingSun//Java//1.0//EN
// Project			:	SPEdit-Prototypes
// ClassName		:	NavigationTree
// Version			:	0.01
// Type				:	Swing custom component
// ThreadSafe		:	Yep
// Description		:	Displays the XMLDocument in a TreeView, allowing to
//						show or hide specific elements.
//						
// URL				:	<http://spedit.sf.net>
// ----------------------------------------------------------------------------
// Authors			:	Sbastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	19-Oct-2001	
// Last mod.		:	19-Oct-2001	
// History			:	
//						19-Oct-2001 First implementation.(SP)
//
// Bugs				:
//						-Does not show selection on the UI.(SP)
//
// To do				:
//						-

package net.sourceforge.spedit.app;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import javax.swing.JTree;
import javax.swing.tree.*;
import javax.swing.event.*;
import net.sourceforge.spedit.core.*;

//-----------------------------------------------------------------------------
//
// NavigationTree 
//
//-----------------------------------------------------------------------------

/**
  * <p>The <code>NavigationTree</code> class allows to represent an
  * <code>XMLDocument</code> as a tree view, allowing to manipulate this
  * <code>XMLDocument</code>.</p>
*/
public class NavigationTree extends JTree implements TreeSelectionListener,
CaretListener
{
	// PROTECTED_______________________________________________________________
	protected Vector		hidden;
	protected Application.DocumentFrame app;
	protected boolean		updateCaret = true;

	//-------------------------------------------------------------------------
	//
	// Constructors
	//
	//-------------------------------------------------------------------------

	public NavigationTree(Application.DocumentFrame app)
	{
		super();
		setRootVisible(false);
		setExpandsSelectedPaths(true);
		setScrollsOnExpand(true);
		this.app = app;
		Font font = new Font("Verdana", Font.PLAIN, 9);
		this.setFont(font);
		hidden = new Vector();
		this.addTreeSelectionListener(this);
		this.app.getEditorPane().addCaretListener(this);
	}

	//-------------------------------------------------------------------------
	//
	// Methods
	//
	//-------------------------------------------------------------------------

	public void setDocument(XMLDocument doc)
	{
		setModel(new Model(doc));
	}

	public TreePath getTreePath(TreeNode node)
	{
		Vector elements = new Vector();
		TreeNode cur = node;

		//We build the parent hierarchy
		while(cur!=null)
		{
			elements.add(cur);
			cur = (TreeNode)cur.getParent();
		}

		//And create the tree path
		TreeNode[] nodes = new TreeNode[elements.size()];
		for ( int i=elements.size()-1 ; i>=0 ; i-- )
		{ nodes[i] = (TreeNode)elements.get(i); }

		return new TreePath(nodes);
	}

	//-------------------------------------------------------------------------
	//
	// Event reaction
	//
	//-------------------------------------------------------------------------

	public void valueChanged(TreeSelectionEvent e) 
	{
		if ( updateCaret )
		{
			XMLNode node = (XMLNode)e.getPath().getLastPathComponent();
			app.getEditorPane().getCaret().moveDot(node.getStartOffset());
			app.getEditorPane().getCaret().setVisible(true);
		}
	}

	public void caretUpdate(CaretEvent e) 
	{
		XMLElementWrapper wrp = (XMLElementWrapper)app.getDocument().getDefaultRootElement();
		TreeNode node = (TreeNode)wrp.getElementAt(e.getDot());
		if ( node!=null )
		{
			updateCaret = false;
			clearSelection();
			TreePath path = getTreePath(node.getParent());
			setSelectionPath(path);
			setLeadSelectionPath(path);
			updateCaret = true;
		}
	}

	//-------------------------------------------------------------------------
	//
	// Tree model
	//
	//-------------------------------------------------------------------------
	
	/**
	  * The <code>Model</code> wraps an <code>XMLDocument</code> allowing to
	  * consider specific nodes as leaves, even if these node are not actually
	  * leaves.
	*/
	protected class Model implements TreeModel
	{
		protected XMLDocument document;

		public Model(XMLDocument doc)
		{
			this.document = doc;
		}

		// CHILDREN_HANDLING__________________________________________________

		public Object getChild(Object parent, int index)
		{
			if ( isLeaf(parent) )
			{ return null; }

			if (parent instanceof XMLElementWrapper)
			{
				int count=0;
				XMLElementWrapper wrp = (XMLElementWrapper)parent;
				//We only count visible nodes.
				for (int i=0 ; i<wrp.getChildCount() ; i++)
				{
					Object child = wrp.getChildAt(i);
					if ( isVisible(child) )
					{	
						if (count==index)
						{	return child;	}
						else
						{	count++;	}
					}
				}
				return null;
			}
			else
			{	return null; }
		}

		public int getChildCount(Object parent) 
		{
			if ( isLeaf(parent) || !isVisible(parent) )
			{ return 0; }
			else if (parent instanceof XMLElementWrapper)
			{
				int count=0;
				XMLElementWrapper wrp = (XMLElementWrapper)parent;
				//We only count visible nodes.
				for (int i=0 ; i<wrp.getChildCount() ; i++)
				{
					if ( isVisible(wrp.getChildAt(i)) )
					{	count++;	}
				}
				return count;
			}
			else
			{ return 0; }
		}

		public int getIndexOfChild(Object parent, Object child) 
		{
			if ( isLeaf(parent) || !isVisible(parent) )
			{	return 0; }
			else if (parent instanceof XMLElementWrapper)
			{
				int count=0;
				XMLElementWrapper wrp = (XMLElementWrapper)parent;
				//We only count visible nodes.
				for (int i=0 ; i<wrp.getChildCount() ; i++)
				{
					if ( isVisible(wrp.getChildAt(i)) )
					{	
						if (wrp.getChildAt(i).equals(child))
						{	return count;	}
						else
						{	count++;	}
					}
				}
				return count;
			}
			else
			{	return 0; }
		}

		public Object getRoot() 
		{
			return document.getDefaultRootElement();
		}

		public boolean isLeaf(Object node) 
		{
			if ( node==null || node instanceof XMLTextWrapper 
				|| hidden.indexOf(node)!=-1 || ((TreeNode)node).isLeaf())
			{	return true;	}
			else
			{	return false;	}
		}

		/**
		  * Tells wether or not the given node is visible inside the tree
		  * model. Name of unvisible nodes are registered into the
		  * <code>NavigationTree</code> <code>hidden</code> vector.
		*/
		public boolean isVisible(Object node) 
		{
			if (node instanceof XMLElementWrapper)
			{
				XMLElementWrapper wrp = (XMLElementWrapper)node;
				if (hidden.indexOf(wrp.getName())==-1)
				{	return true; }
				else
				{	return false;	}
			}
			else 
			{	return false; }
		}

		// CHANGE_HANDLING_____________________________________________________

		public void addTreeModelListener(TreeModelListener l) 
		{
			document.addTreeModelListener(l);
		}

		public void removeTreeModelListener(TreeModelListener l) 
		{
			document.removeTreeModelListener(l);
		}

		public void valueForPathChanged(TreePath path, Object newValue) 
		{
			document.valueForPathChanged(path, newValue);
		}
	}

}

// EOF-Linux/ISO-8859-1-------------------------------@RisingSun//Java//1.0//EN
