// ---------------------------------------------------@RisingSun//Java//1.0//EN
// Project			:	SPEdit-Prototypes
// ClassName		:	Styleable
// Version			:	0.98
// Type				:	Core interface
//						
// ThreadSafe		:	Nope.
// Description		:	Describes the methods of an object that can store
//						information about its style.
// URL				:	<http://SPEdit.sourceforge.net>
// ----------------------------------------------------------------------------
// Authors			:	Sbastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	25-Sep-2001	
// Last mod.		:	09-Nov-2001	
// History			:	
//						09-Nov-2001 Added border style attributes.(sp)
//						29-Oct-2001 Added title and title padding
//						support.(sp)
//						22-Oct-2001 Better initialization, padding as
//						shorts.(sp)
//						18-Oct-2001 Merged style attributes initialisation from
//						the zombie TextView class.(SP)
//						02-Oct-2001 Fixed small bug in style inheritance: new
//						inherited stylesheet did not mask every other inherited
//						style.(SP)
//						01-Oct-2001 Implementation of PropertyChange listener,
//						yet another restructuration of the Styleable
//						interface.(SP)
//						28-Sep-2001 Implementation of a default style
//						object. (SP)
//						27-Sep-2001 Major redesign of the stylable
//						interface.(SP)
//						26-Sep-2001 Utils class.(SP)
//						25-Sep-2001 First implementation.(SP)
//
// Bugs				:
//						-
//
// To do			:
//						- Add property change listening on inherited
//						stylesheets.

package net.sourceforge.spedit.core;

import java.awt.Color;
import java.awt.Font;
import java.util.*;
import java.beans.*;
import gnu.regexp.*;

/**
  * <p>The <code>Styleable</code> interface describes the methods that a component
  * should implement to allow style customization.</p>
  * <p>This interface allows unlimited style properties for a given object.
  * Moreover those styles handle dynamic, selective inheritance. This mechanism
  * is much more powerful than CSS, as the styles are also organized
  * hierarchically.</p>
  * <p>The organizational model of styles follows in a way the package
  * organisation of Java. Each style property has an identifier which is
  * composed of words, including dashes, numbers and underscores, separated by
  * dots. For instance the <code>font.name</code> property means the
  * <code>name</code> property of the <code>font</code> style group.</p>
  * <p>This mechanism allows to inherit only parts of a given style (eg. only
  * the font size for instance).</p>
  * 
  * @version 0.98
  * @author Sbastien Pierre
*/
public interface Styleable
{

	public abstract StyleSheet getStyleSheet();
	public abstract void setStyleSheet(StyleSheet style);
	public abstract StyleSheet getParentStyleSheet();

	//------------------------------------------------------------------------
	//
	// StyleSheet
	//
	//------------------------------------------------------------------------

	/**
	  * <p>The <code>StyleSheet</code> provides a flexible way to store
	  * structured, hierarchized style information. These stylesheet support
	  * inheritance of whole, or part of the style tree structure, plus
	  * resolution into a parent <code>Styleable</code> element.</p>
	*/
	public static class StyleSheet implements PropertyChangeListener
	{
		// PUBLIC____________________________________________________________

		public static final float	LEFT=0f;
		public static final float	CENTER=0.5f;
		public static final float	RIGHT=1f;

		// PROTECTED_________________________________________________________

		/**
		  * This hash stores the different root properties and property groups.
		  * For example <code>font.name</code> and <code>font.size</code> will
		  * be accessed by getting the <code>font</code> entry of the has,
		  * which will be a hash itself and which will contain the
		  * <code>name</code> and <code>size</code> properties.
		*/
		protected String    name;
		protected Hashtable properties;
		protected Hashtable propertiesClasses;
		protected Hashtable inherited;
		protected Vector    listeners;
		protected Styleable parent;

		//Style properties
		public short		paddingLeft;
		public short		paddingRight;
		public short		paddingTop;
		public short		paddingBottom;
		public int			lineHeight;
		public float		align;
		public float		width;
		public Font			font;
		
		public Color		foreground;
		public Color		background;
		
		public String		title;
		public Font			titleFont;
		public short		titlePaddingLeft;
		public short		titlePaddingRight;
		public short		titlePaddingTop;
		public short		titlePaddingBottom;
		
		public int			borderSize;
		public Color		borderColor;
		public short		borderPaddingLeft;
		public short		borderPaddingRight;
		public short		borderPaddingTop;
		public short		borderPaddingBottom;
		
		public boolean		hasExpander;

		//---------------------------------------------------------------------
		//
		// Constructors
		//
		//---------------------------------------------------------------------


		public StyleSheet()
		{
			this("Unnamed");
		}
	
		public StyleSheet(String name)
		{
			this(name,null);
		}

		public StyleSheet(String name, Styleable parent)
		{
			this.name = name;
			this.parent = parent;
			init();
		}

		protected void init()
		{
			properties = new Hashtable();
			propertiesClasses = new Hashtable();
			inherited  =  new Hashtable();
			listeners = new Vector(2);
			setParent(parent);
		}

		// METHODS___________________________________________________________

		//---------------------------------------------------------------------
		//
		// Style management
		//
		//---------------------------------------------------------------------

		/**
		  * Specifies that the given property has to be of the given class.
		*/
		public void setStyleClass(String property, Class pclass)
		{
			Class mclass = (Class)propertiesClasses.get(property);
			//Has the property type already been defined?
			if ( mclass==null )
			{
				//We check that the property has not been inherited
				StyleSheet owner = getStyleOwner(property);
				if ( owner==this || owner==null)
				{
					propertiesClasses.put(property, pclass);
				}
			}
		}

		/**
		  * Sets the style property designated by the <code>property</code>
		  * parameter to the given <code>value</code>.
		*/
		public void setStyle(String property, Object value)
		{
			Class pclass = (Class)propertiesClasses.get(property);
			Object old = properties.get(property);
			properties.put(property, value);
			if ( pclass==null )
			{ propertiesClasses.put(property, value.getClass());}
			firePropertyChange( new PropertyChangeEvent(this, property, old,
			value));
		}

		/**
		  * Retrieves the parent of the given property. It can be the
		  * <code>this</code> object in case the property is defined locally,
		  * otherwise the parent from which the property is inherited is given.
		  * In case the property is not defined nor inherited,
		  * <code>null</code> is returned.
		*/
		protected StyleSheet getStyleOwner(String property)
		{
			if ( properties.get(property)!=null )
			{	return this;	}
			else
			{
				String[] p = Utils.splitStyle(property);
				for( int last=p.length ; last>=0 ; last-- )
				{
					StyleSheet parent = (StyleSheet)inherited.get(genStyleName(p,last));
					if (parent!=null)
					{	return parent;	}
				}
				//The last resort
				StyleSheet pss = (StyleSheet)inherited.get(".");
				if (pss!=null)
				{	return pss;	}
				else if ( parent!=null )
				{
					return parent.getParentStyleSheet();
				}
			}
			return null;
		}

		/**
		  * Generates a part of the style name from the given style
		  * subcomponents. For example if property is <code>[ "block",
		  * "border", "type" ]</code> and <code>last</code> is 2, the result
		  * will be <code>block.border</code>.
		*/
		private String genStyleName(String[] property, int last)
		{
			String res=".";
			if ( last>property.length )
			{	last = property.length;	}
			for ( int i=0 ; i<last ; i++ )
			{
				if ( i==0 )
				{	res=property[0];	}
				else
				{	res +="."+property[1];	}
			}
			return res;
		}

		/**
		  * Gets the given style propery. This can be either a list or a style
		  * object. In case the returned object is a List this will mean the given
		  * property is a property set.
		*/
		public Object getStyle(String property)
		{
			Object obj = properties.get(property);
			//If the property is not defined locally
			if ( obj==null )
			{
				StyleSheet owner = getStyleOwner(property);
				//We look into the inherited properties
				if ( owner!=null )
				{
					return owner.getStyle(property);
				}
				//Or into the Styleable parent if it is defined.
				else
				{	
					if (parent!=null)
					{
						StyleSheet ss = parent.getParentStyleSheet();
						if (ss!=null)
						{	return ss.getStyle(property); }
						else
						{	return null; }
					}
					else
					{	return null; }
				}
			}
			else
			{
				return obj;
			}
		}

		/**
		  * Returns the class of the object bound to the given style property. This
		  * is useful if you want to check what type of object you can put inside a
		  * property.
		*/
		public Class getStyleClass(String property)
		{
			Class cl = (Class)propertiesClasses.get(property);
			if ( cl==null )
			{
				StyleSheet owner = getStyleOwner(property);
				if ( owner!=null )
				{ return owner.getStyleClass(property); }
				else
				{ return null; }
			}
			else
			{
				return cl;
			}
		}

		/**
		  * Inherits all the properties of the given stylable object.
		*/
		//TODO: Maybe should fire a property change event(sp).
		public void inheritStyle(StyleSheet stylesheet)
		{
			for (Enumeration e=inherited.elements() ; e.hasMoreElements() ; )
			{
				((StyleSheet)e.nextElement()).removePropertyChangeListener(this);
			}
			inherited.clear();
			inherited.put(".", stylesheet);
			stylesheet.addPropertyChangeListener(this);
			update();
		}

		/**
		  * Inherits only a given property or group of properties from the given
		  * stylable object.
		*/
		public void inheritStyle(String property, StyleSheet stylesheet)
		{
			inherited.put(property, stylesheet);
			stylesheet.addPropertyChangeListener(this);
			update();
		}

		public Enumeration listProperties(String group)
		{
			Vector result = new Vector();
			Enumeration e;

			//We merge the local properties
			if (group.equals("."))
			{
				for (e=properties.keys() ; e.hasMoreElements() ; )
				{	result.add(e.nextElement());	}
			}
			else
			{
				for (e=properties.keys() ; e.hasMoreElements() ; )
				{	
					String key = (String)e.nextElement();
					if ( key.indexOf(group)==0 )
					{	result.add(key);	}
				}
			}

			//We merge the inherited properties
			for (e=inherited.keys() ; e.hasMoreElements() ; )
			{
				String key = (String)e.nextElement();
				Enumeration e2 = ((StyleSheet)inherited.get(key)).listProperties(key);
				while ( e2.hasMoreElements() )
				{
					Object el = e2.nextElement();
					if (!result.contains(el))
					{	result.add(el); }
				}
			}
			return result.elements();
		}

		//---------------------------------------------------------------------
		//
		// Properties
		//
		//---------------------------------------------------------------------

		/**
		  * Adds a listener that will be called each time a property of the
		  * stylable object changes.
		*/
		public void addPropertyChangeListener(PropertyChangeListener l)
		{
			//We make sure the given listener has not already been added.
			if ( listeners.indexOf(l)==-1 )
			{	listeners.add(l); }
		}

		public void removePropertyChangeListener(PropertyChangeListener l)
		{
			listeners.remove(l);
		}

		public void propertyChange(PropertyChangeEvent evt) 
		{
			update();
			firePropertyChange(evt);
		}

		protected void firePropertyChange(PropertyChangeEvent evt) 
		{
			ListIterator it = listeners.listIterator();
			while ( it.hasNext() )
			{
				PropertyChangeListener l=(PropertyChangeListener)it.next(); 
				l.propertyChange(evt);
			}
		}

		//---------------------------------------------------------------------
		//
		// Attributes
		//
		//---------------------------------------------------------------------


		public String getName()
		{
			return name;
		}

		public void setName(String name)
		{
			this.name = name;
		}
		
		public Styleable getParent()
		{
			return parent;
		}

		public void setParent(Styleable parent)
		{
			StyleSheet sheet;

			//Removes property change listener from the old parent
			if (this.parent!=null)
			{
				sheet = this.parent.getParentStyleSheet();
				if ( sheet!=null )
				{ sheet.removePropertyChangeListener(this); }
			}

			this.parent = parent;

			//Register as property change listener
			if (parent!=null)
			{
				sheet = parent.getParentStyleSheet();
				if ( sheet!=null )
				{ sheet.addPropertyChangeListener(this); }
			}

			//Then update
			update();
		}

		//---------------------------------------------------------------------
		//
		// Updating
		//
		//---------------------------------------------------------------------

		int counter=0;
		public void update()
		{
			//Foreground color
			if (getStyle("color.foreground")==null)
			{ foreground = Color.black; }
			else
			{ foreground = (Color)getStyle("color.foreground"); }

			//Background color
			if (getStyle("color.background")==null)
			{ background = null; }
			else
			{ background = (Color)getStyle("color.background"); }

			//Left padding
			if (getStyle("padding.left")==null)
			{ paddingLeft = (short)0; }
			else
			{ paddingLeft = ((Number)getStyle("padding.left")).shortValue(); }

			//Right padding
			if (getStyle("padding.right")==null)
			{ paddingRight = (short)0; }
			else
			{ paddingRight = ((Number)getStyle("padding.right")).shortValue(); }

			//Top padding
			if (getStyle("padding.top")==null)
			{ paddingTop = (short)0; }
			else
			{ paddingTop = ((Number)getStyle("padding.top")).shortValue(); }

			//Bottom padding
			if (getStyle("padding.bottom")==null)
			{ paddingBottom = (short)0; }
			else
			{ paddingBottom = ((Number)getStyle("padding.bottom")).shortValue(); }

			//Font size
			int fontsize=12;
			if (getStyle("font.size")!=null)
			{ fontsize = ((Number)getStyle("font.size")).intValue(); }

			//Font type
			int fonttype=Font.PLAIN;
			if (getStyle("font.type")!=null)
			{ 
				String type = (String)getStyle("font.type");
				type = type.toLowerCase();
				if ( type.equals("bold") )
				{	fonttype = Font.BOLD;	}
				else if ( type.equals("italic") )
				{	fonttype = Font.ITALIC;	}
				else
				{	fonttype = Font.PLAIN;	}
			}

			//Font name
			String fontname="Times";
			if (getStyle("font.name")!=null)
			{ fontname = (String)getStyle("font.name"); }
			font = new Font(fontname, fonttype, fontsize);

			//Title Font size
			if (getStyle("title.font.size")!=null)
			{ fontsize = ((Number)getStyle("title.font.size")).intValue(); }

			//Title Font size
			if (getStyle("title.content")!=null)
			{ this.title = ((String)getStyle("title.content")); }
			else
			{ this.title = null; }

			//Title Font type
			if (getStyle("title.font.type")!=null)
			{ 
				String type = (String)getStyle("title.font.type");
				type = type.toLowerCase();
				if ( type.equals("bold") )
				{	fonttype = Font.BOLD;	}
				else if ( type.equals("italic") )
				{	fonttype = Font.ITALIC;	}
				else
				{	fonttype = Font.PLAIN;	}
			}

			//Title font name
			if (getStyle("title.font.name")!=null)
			{ fontname = (String)getStyle("title.font.name"); }
			titleFont = new Font(fontname, fonttype, fontsize);

			//Title left padding
			if (getStyle("title.padding.left")==null)
			{ titlePaddingLeft = (short)0; }
			else
			{ titlePaddingLeft = ((Number)getStyle("title.padding.left")).shortValue(); }

			//Title right padding
			if (getStyle("title.padding.right")==null)
			{ titlePaddingRight = (short)0; }
			else
			{ titlePaddingRight = ((Number)getStyle("title.padding.right")).shortValue(); }

			//Title top padding
			if (getStyle("title.padding.top")==null)
			{ titlePaddingTop = (short)0; }
			else
			{ titlePaddingTop = ((Number)getStyle("title.padding.top")).shortValue(); }

			//Title bottom padding
			if (getStyle("title.padding.bottom")==null)
			{ titlePaddingBottom = (short)0; }
			else
			{ titlePaddingBottom = ((Number)getStyle("title.padding.bottom")).shortValue(); }

			//Line height
			if (getStyle("font.line-height")==null)
			{ lineHeight = fontsize+2; }
			else
			{ lineHeight = ((Number)getStyle("font.line-height")).intValue(); }

			//Block width
			if (getStyle("block.width")==null)
			{ width = 400; }
			else
			{ width = ((Number)getStyle("block.width")).floatValue(); }

			//Block alignment
			if (getStyle("block.align")==null)
			{ align = LEFT; }
			else
			{ 
				String a = (String)getStyle("block.align");
				a=a.toLowerCase();
				if ( a.equals("center") )
				{ align = CENTER;	}
				else if ( a.equals("right") )
				{ align = RIGHT;	}
				else if ( a.equals("left") )
				{ align = LEFT;	}
				else
				{ align = Float.parseFloat(a); }
			}

			//Border size
			if (getStyle("border.size")==null)
			{ borderSize = 0; }
			else
			{ borderSize = ((Number)getStyle("border.size")).intValue(); }

			//Border color
			if (getStyle("border.color")==null)
			{ borderColor = Color.black; }
			else
			{ borderColor = (Color)getStyle("border.color"); }
			
			//Border left padding
			if (getStyle("border.padding.left")==null)
			{ borderPaddingLeft = (short)0; }
			else
			{ borderPaddingLeft = ((Number)getStyle("border.padding.left")).shortValue(); }
			
			//Border bottom padding
			if (getStyle("border.padding.right")==null)
			{ borderPaddingRight = (short)0; }
			else
			{ borderPaddingRight = ((Number)getStyle("border.padding.right")).shortValue(); }

			//Border top padding
			if (getStyle("border.padding.top")==null)
			{ borderPaddingTop = (short)0; }
			else
			{ borderPaddingTop = ((Number)getStyle("border.padding.top")).shortValue(); }
			
			//Border bottom padding
			if (getStyle("border.padding.bottom")==null)
			{ borderPaddingBottom = (short)0; }
			else
			{ borderPaddingBottom = ((Number)getStyle("border.padding.bottom")).shortValue(); }
			
			//Expander
			if (getStyle("expander.presence")==null)
			{ hasExpander=false; }
			else
			{ hasExpander=true;; }
			
			counter++;
		}

		//---------------------------------------------------------------------
		//
		// Utils
		//
		//---------------------------------------------------------------------

		public String toString()
		{
			StringBuffer result = new StringBuffer();
			result.append("StyleSheet: "+name+"\n");
			Vector parents = new Vector();
			for (Enumeration e=listProperties(".") ; e.hasMoreElements() ;)
			{
				String prop=(String)e.nextElement();
				result.append("  "+prop+" = "+getStyle(prop)+"\n");
			}
			if ( parent!=null && parent.getParentStyleSheet()!=null )
			{
				result.append("\n from parent "+
				parent.getParentStyleSheet().toString());
			}
			else
			{
				if ( parent==null )
				{	result.append("\n no parent.");}
				else
				{	result.append("\n no parent stylesheet.");}
			}
			return result.toString();
		}
	}

	//------------------------------------------------------------------------
	//
	// Utils
	//
	//------------------------------------------------------------------------

	public static class Utils
	{
		// PUBLIC_____________________________________________________________
		public static final int COLOR = 1;
		public static final int PIXEL = 2;
		public static final int POINT = 3;
		public static final int UNKNOWN= 9;

		// METHODS____________________________________________________________

		/**
		  * Identifies the given string, telling what kind of content it holds
		  * (either COLOR, PIXEL, POINT or UNKNOWN)
		*/
		public static int identify(String string)
		{
			RE RCOLOR=null,RPIXEL=null,RPOINT=null;
			try
			{
				RCOLOR = new RE("#(([0-9a-fA-F]{6})([0-9a-fA-F]{2})?)$");
				RPIXEL = new RE("(\\d+)$");
				RPOINT = new RE("((\\d+)(\\.\\d+))?pt$");
			}
			catch (Exception e)
			{ e.printStackTrace(); }

			if ( RCOLOR.getMatch(string)!=null )
			{	return COLOR;	}
			else if ( RPIXEL.getMatch(string)!=null )
			{	return PIXEL;	}
			else if ( RPOINT.getMatch(string)!=null )
			{	return POINT;	}
			else
			{	return UNKNOWN;	}
		}

		/**
		  * Parses the given string and returns the corresponding color. The
		  * string must match the RCOLOR regular expression.
		  * @param <code>string</code> represents the string in the following
		  * form: <code>#RRGGBB</code> or <code>#RRGGBBAA</code> in case you
		  * want an alpha channel.
		*/
		public static Color parseColor(String string)
		{
			RE RCOLOR=null;

			try
			{
				RCOLOR = new RE("#(([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})?)$");
			}
			catch (Exception e)
			{ e.printStackTrace(); }

			int red, green, blue, alpha=255;
			REMatch match = RCOLOR.getMatch(string);
			red   = fromHexa(match.toString(2));
			green = fromHexa(match.toString(3));
			blue  = fromHexa(match.toString(4));
			//Does the color has an alpha channel ?
			if ( match.getEndIndex(5)!= -1 )
			{ alpha = fromHexa(match.toString(5)); }
			return new Color(red, green, blue, alpha);
		}

		protected static int fromHexa(String s)
		{
			char b = s.charAt(1);
			char a = s.charAt(0);
			int  res = 0;

			if ( b>='a' && b<='f' )
			{	res += 10+b-'a';	}
			else if ( b>='A' && b<='F' )
			{	res += 10+b-'A';	}
			else if ( b>='0' && b<='9' )
			{	res += b-'0';	}

			if ( a>='a' && a<='f' )
			{	res += (10+a-'a')*16;	}
			else if ( a>='A' && a<='F' )
			{	res += (10+a-'A')*16;	}
			else if ( a>='0' && a<='9' )
			{	res += (a-'0')*16;	}

			return res;
		}

		public static double parsePoint(String string)
		{
			RE RPOINT=null;
			try
			{
				RPOINT = new RE("^((\\d)+(\\.\\d+))?pt$");
			}
			catch (Exception e)
			{ e.printStackTrace(); }

			REMatch match = RPOINT.getMatch(string);
			return Double.parseDouble(match.toString(1));
		}

		public static long parsePixel(String string)
		{
			return Long.parseLong(string);
		}

		/**
		  * Takes a property style name, like <code>block.border.type</code>
		  * and will return an array of string composed of the different
		  * groups. In this case the result will be <code>["block", "border",
		  * "type"].
		*/
		public static String[] splitStyle(String style)
		{
			int i=0, start=0, length=1;
			while ( (i=style.indexOf('.', start))!=-1 )
			{
				length++;
				start = i+1;
			
			}
			String[] result = new String[length];
			start = i = length = 0;
			while ( (i=style.indexOf('.', start))!=-1 )
			{
				result[length]=style.substring(start,i);
				length++;
				start = i+1;
			}
			result[length] = style.substring(start, style.length());
			return result;
		}
	}
}
