// ---------------------------------------------------@RisingSun//Java//1.0//EN
// Project			:	SPEdit-Prototypes
// ClassName		:	XMLDocument
// Version			:	0.80
// Type				:	Hybrid JDOM/Swing document with TreeModel interface.
//
// ThreadSafe		:	No
// Description		:	Wraps around an XML document (as of org.jdom package)
// 						and implements the javax.swing.text.Document interface.
// URL				:	<http://SPEdit.sourceforge.net>
// ----------------------------------------------------------------------------
// Authors			:	Sbastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	03-Sep-2001	
// Last mod.		:	07-Nov-2001	
// History			:
//						07-Nov-2001 Improved document filtering with
//						strippin.(sp)
//						25-Oct-2001 Added some javadoc.(sp)
//						17-Oct-2001 Event management revamp.(SP)
//						16-Oct-2001 JavaDoc, bugfixes.(SP)
//						15-Oct-2001 Moved to StyledDocument to benefit from the
//						views that only work with such kind of documents.(SP)
//						27-Sep-2001 Added the filter method.(SP)
//						24-Sep-2001 Redone document event managment.(SP)
//						20-Sep-2001 Implemented parts of the event managment
//						system.(SP)
//						17-Sep-2001 Made the loadFromFile static.(SP)
//						14-Sep-2001 Encapsulating the XMLElement text
//						functions. Loads an XML file as a document.(SP)
//						10-Sep-2001 Improved Javadoc. Move XMLAttributeSet in
//						its own class.(SP)
//						07-Sep-2001 Implentation of the mixed JDOM/Swing model.
//						05-Sep-2001 Implementation of org.jdom interfaces.(SP)
//						04-Sep-2001 Element offset localisation.(SP)
//						03-Sep-2001 First implementation.(SP)
//
// Notes			:	Some portions of code have been taken from the JDK and
//						adapted to the current contex.(SP)
// Bugs				:
//					  -
//
// To do			:
//					  - Support of Undo/Redo.
//					  - Understand the role of the CHANGE event.
//					  - Handle position biases
//					  - The document should store an array containing all the
//					    text in the document, in order to prevent the 
//					  - Event handling in Swing DOM and TreeModel
//					  - Improved the EventHandling model (speed issues).
//					  - In insertString, implement attributes merging.
//					  - Implement the write locking.
//					  - Errors when there is no root node in most methods.
//					  - Optimize event handling and position updates.

package net.sourceforge.spedit.core;

import java.io.*;
import java.util.*;
import java.text.*;
import java.awt.Color;
import java.awt.Font;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.undo.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.adapters.*;

//-----------------------------------------------------------------------------
//
// XMLDocument class 
//
//-----------------------------------------------------------------------------

/**
  * <p>The <code>XMLDocument</code> class wraps a Swing <code>Document</code>
  * around a JDOM XML document. There is an article on the website that
  * describes how the design of this class was made, for further understanding
  * you should read the source code.</p>
  * <p>When using the <code>XMLDocument</code> API you should take great care
  * of the objects that you manipulate. Here are some points that should be
  * clear:
  * <ul>
  *   <li>Everything that is returned are <em>wrapper</em> objects around JDOM
  *   objects. Those wrappers implement the <code>XMLNode</code> interface,
  *   there are currently only two types of wrappers :
  *   <code>XMLElementWrapper</code> and <code>XMLTextWrapper</code>, which
  *   respectively wrap JDOM <code>Element</code> and <code>Text</code>
  *   nodes.</li>
  *   <li>All wrappers also implement the <code>TreeNode</code> interface in
  *   order to ease representation as in a <code>JTree</code>. This can lead to
  *   some confusion as some methods may seem redundant. Do not worry in this
  *   case sa you will always be able to cast to an <code>XMLNode</code>
  *   interface.</li>
  * </ul>
  * @author Sbastien Pierre
  * @version 0.75
*/
public class XMLDocument extends org.jdom.Document implements StyledDocument,
			 TreeModel
{
	// PUBLIC__________________________________________________________________
	public static final String PARSER="net.sourceforge.spedit.core.XMLParser";
	public static final SAXBuilder saxBuilder = new SAXBuilder(PARSER);
	public static final XMLOutputter xmlOutputter = new XMLOutputter();
	public static final JDOMFactory xmlFactory= new XMLJDOMFactory();

	// PROTECTED_______________________________________________________________
	protected static final int POSITIONS_THRESHOLD=10;
	protected static final int GAP_LENGTH=10;
	protected Element[]   roots = new Element[1];
	protected Vector      documentListeners;
	protected Vector      undoableEditListeners;
	protected Vector      treeModelListeners;
	protected Hashtable   properties;
	protected List        positions;
	protected GapContent  content;
	protected javax.swing.text.Position    startPosition;
	protected javax.swing.text.Position    endPosition;
	protected Font        font;

	// PRIVATE_________________________________________________________________

	//-------------------------------------------------------------------------
	//
	// Constructors
	//
	//-------------------------------------------------------------------------

	/**
	  * Creates a new blank <code>XMLDocument</code>.
	*/
	public XMLDocument()
	{
		super();
		init();
	}

	/**
	  * Creates a new <code>XMLDocument</code> with the given
	  * <code>rootElement</code> as a root node.
	  * @param <code>rootElement</code> the root node of this
	  * <code>XMLDocument</code>.
	*/
	public XMLDocument(org.jdom.Element rootElement)
	{
		super(rootElement);
		init();
	}

	/**
	  * <p>Creates a new <code>XMLDocument</code> with the given
	  * <code>rootElement</code> as a root node, with the specified document
	  * type. This constructor is not really useful, you should not use it as
	  * no consistency check is made with the given doctype.</p>
	  * <p>This constructor is provided for <code>JDOMFactory</code>
	  * consistency purposes.</p>
	  * @param <code>rootElement</code> the root node of this
	  * 	   <code>XMLDocument</code>.
	  * @param <code>docType</code> the root node of this
	  * 	   <code>XMLDocument</code>.
	*/
	public XMLDocument(org.jdom.Element rootElement, org.jdom.DocType docType)
	{
		super(rootElement, docType);
		init();
	}

	/**
	  * Initializes the <code>XMLDocument</code>. It should not be called after
	  * document creation. This method is present to provide a common
	  * initialization facility for the various constructors, which is a
	  * technique cleaner that <code>this(...)</code> calls, as I allows
	  * <code>super(...)</code> constructor calls with more paramters.
	*/
	protected void init()
	{
		properties = new Hashtable();
		documentListeners = new Vector();
		undoableEditListeners = new Vector();
		treeModelListeners = new Vector();
		positions = new ArrayList(POSITIONS_THRESHOLD);
		font = new Font("Times",Font.PLAIN, 12);

		startPosition = new Position(0, javax.swing.text.Position.Bias.Backward);
		endPosition = new Position(0, javax.swing.text.Position.Bias.Forward);

		saxBuilder.setFactory(xmlFactory);

		//We create the content
		content = new GapContent(GAP_LENGTH);
		try
		{
			if ( getDefaultRootElement()!=null )
			{
				content.insertString(0,
				((XMLElementWrapper)getDefaultRootElement()).getWholeText());
			}
		}
		catch (Exception e)
		{ e.printStackTrace(); }
	}

	//-------------------------------------------------------------------------
	//
	// XML facilities
	//
	//-------------------------------------------------------------------------
	
	// DOCUMENT_LOADING________________________________________________________

	/**
	  * <p>Loads the given file and tries to generate an
	  * <code>XMLDocument</code> instance from its content. This method returns
	  * a new <code>XMLDocument</code> in case of success, or of course
	  * </code>null</code> in case of failure.</p> <p>If an exception occurs,
	  * it will be dumped on <em>stderr</em>.</p>
	  * @param file Is the XML source file to be loaded
	  * @param fact Is the <code>XMLViewFactory</code> used to filter the element
	  * contents. It can be set to null.
	*/
	public static XMLDocument loadFromFile(java.io.File file, XMLViewFactory fact)
	{
		try
		{
			saxBuilder.setFactory(xmlFactory);
			XMLDocument doc = (XMLDocument)saxBuilder.build(file);
			doc.filter(fact);
			doc.content.insertString(0,
			((XMLElementWrapper)doc.getDefaultRootElement()).getWholeText());
			
			return doc;
		}
		catch(BadLocationException e)
		{
			e.printStackTrace(System.err);
			return null;
		}
		catch(org.jdom.JDOMException e)
		{
			System.err.println("Unable to load the source XML file.");
			System.err.println("  "+e.toString());
			e.printStackTrace(System.err);
			return null;
		}
	}
	
	// DOCUMENT_SAVING__________________________________________________________

	/**
	  * <p>Saves the current <code>XMLDocument</code> into the given file. If
	  * an exception it will be thrown.</p>
	*/
	public void saveToFile(java.io.File file)
	throws IOException
	{
		java.io.OutputStream os=null;
		try
		{
			os = new java.io.FileOutputStream(file);
			xmlOutputter.output(convertToJDOM(), os);
		}
		finally
		{ if (os!=null) { os.close(); } }
	}

	// JDOM_CONVERSION__________________________________________________________

	/**
	  * <p>Generates a natural JDOM document. A <i>natural</i> JDOM document means
	  * a document that does not have the fancy elements, such as
	  * <code>XMLText</code> elements as children. Actually an XMLDocument can
	  * not really be used "as is" with the JDOM facilities, you first have to
	  * convert it using this method.</p>
	  * <p>Note that if you plan to customize the <code>XMLElement</code>
	  * content you should extend the <code>convertElement</code>
	  * behaviour.</p>
	  * @return A natural JDOM copy of this document. The new document will
	  * have completely different element references, so you don't have to
	  * worry about changes in the current document.
	  * @see #convertElement
	*/
	public org.jdom.Document convertToJDOM()
	{
		return new org.jdom.Document(convertElement(getRootElement()));
	}

	/**
	  * Converts the given element into a natural JDOM element. This methods
	  * actually replaces <code>XMLText</code> elements by strings added to the
	  * new element using the <code>addContent</code> method of the
	  * <code>org.jdom.Element</code>.
	  * @param <code>e</code> The elements that will serve as a clone model for
	  * the return value.
	  * @result The conversion to a natural JDOM element of the given element.
	*/
	protected org.jdom.Element convertElement(org.jdom.Element e)
	{
		org.jdom.Element elt = new org.jdom.Element(e.getName(), e.getNamespace());
		elt.setAttributes(e.getAttributes());
		for ( int i=0 ; i<e.getContent().size() ; i++ )
		{
			Object child = e.getContent().get(i);
			if ( child instanceof XMLText )
			{
				elt.addContent(((XMLText)child).toString());
			}
			else if ( child instanceof org.jdom.Element )
			{
				elt.addContent(convertElement((org.jdom.Element)child));
			}
		}
		return elt;
	}

	// DOCUMENT_COSMETICS_______________________________________________________

	/**
	  * <p>Filters the given document, removing text nodes that contain no
	  * characters. Those strings are always present inside an XML document
	  * that you have just loaded. This method should only be called when the
	  * document content has been (re)set.</p>
	  * <p>In addition to that the filter looks in the given <b>factory</b> for
	  * information on how to filter element textual content, which can be for
	  * example stripping of leading and trailing spaces.</p>
	  * @param factory The factory from which the information will be extracted.
	*/
	public void filter(XMLViewFactory factory)
	{
		filter(getRootElement(), factory);
	}

	/**
	  * Filters the given element recursively. Note that this method is
	  * desctructive, so you should not use it without being sure you do not
	  * remove valuable information.
	  * @param element the element to filter.
	  * @param factory the factory from which filtering information can be
	  *		   found.
	*/
	protected void filter(org.jdom.Element element,XMLViewFactory factory)
	{
		Vector to_be_detached = new Vector();
		Vector new_content    = new Vector();
		XMLInformation info	  = null;
		if ( factory!=null )
		{ 
			info = factory.getElementInformation(((XMLElement)element).getWrapper());
		}
		
		//We look for the insertion point
		for ( int i=0 ; i<element.getContent().size() ; i++ )
		{
			Object obj = element.getContent().get(i);
			//It is an Element, so we recurse
			if ( obj instanceof XMLElement )
			{
				XMLElement nelt = (XMLElement)obj;
				filter(nelt,factory);
				new_content.add(nelt);
				to_be_detached.add(nelt);
			}
			else if ( obj instanceof XMLText )
			{
				XMLText nelt = (XMLText)obj;
				if ( !isGarbage(nelt) )
				{	
					if (info!=null)
					{
						if (info.hasStrip())
						{	nelt.strip();	}
					}
					new_content.add(nelt);
				}
			}
		}
		//Now we detach the nodes
		for (Enumeration e=to_be_detached.elements();e.hasMoreElements();)
		{ ((XMLElement)e.nextElement()).detach(); }

		//We set the new content
		((XMLElement)element).setNewContent(new_content);
	}

	/**
	  * Tells wether or not the given element is a garbage element. A garbage
	  * element is typically an element that only contains blank characters.
	  * @param <code>elt</code> The <code>XMLText</code> element to analyze.
	  * This only applies to this kind of elements as they are the only
	  * susceptible to contain only text in the <code>XMLDocument</code> model.
	*/
	protected boolean isGarbage(XMLText elt)
	{
		if ( elt.getValue().trim().length()>0 )
		{	return false;	}
		else
		{	return true;	}
	}
	//-------------------------------------------------------------------------
	//
	// Swing Document model compliance
	//
	//-------------------------------------------------------------------------

	// ELEMENTS_MANIPULATION___________________________________________________

	/**
	  * Prepends the element at the given offset in the text.
	  * @param <code>element</code> is the element that will be inserted.
	  * @param <code>offset</code> is the offset of the insertion point in the
	  * text.
	*/
	public void insertElement(XMLElement element, int offset)
	{
		XMLElement root = (XMLElement)super.getRootElement();
		root.getWrapper().insert(element, offset, javax.swing.text.Position.Bias.Backward);
	}

	/**
	  * Returns the default root element of the document. In the particular
	  * case of an <code>XMLDocument</code>, there is a always a single root.
	*/
	public Element getDefaultRootElement()
	{
		XMLElement root = (XMLElement)super.getRootElement();
		if (root==null)
		{	return null;	}
		return root.getWrapper();
	}

	/**
	  * Returns an array of elements which only contains the default document
	  * root.
	*/
	public Element[] getRootElements()
	{
		XMLElement root = (XMLElement)super.getRootElement();
		if (root==null)
		{	return null;	}
		roots[0] = getDefaultRootElement();
		return roots;
	}

	public org.jdom.Document setRootElement(org.jdom.Element elem)
	{
		if (elem==null)
		{	return this;	}

		super.setRootElement(elem);
		return this;
	}

	/**
	  * Returns the <code>XMLElement</code> that contains the specified offset.
	  * The name of this method is a little bit abusive, as the actual
	  * <code>XMLElement</code> may not a be a <em>paragraph</em> element, but
	  * this is provided to conform to the Swing document interface.
	*/
	public Element getParagraphElement(int offset)
	{
		XMLElement root = (XMLElement)super.getRootElement();
		if (root==null)
		{	return null;	}
		Element e = root.getWrapper().getElementAt(offset);
		if ( e.isLeaf() )
			return e.getParentElement();
		else
			return e;
	}

	/**
	  * Search the Swing Element which is located at the specified position.
	  * If the position is negative or superior to the document text length,
	  * then null is returned.
	  * @param	<code>offset</code> The offset in the text, starting at 0, ending
	  * at <code>document.getLength()</code>, exclusive.
	  * @return The searched <code>Element</code> or null if the position is
	  * out of the document bounds.
	*/
	public Element getElementAt(int offset)
	{
		XMLElement root = (XMLElement)super.getRootElement();
		return root.getWrapper().getElementAt(offset);
	}

	// LISTENERS_______________________________________________________________
	public void addDocumentListener(DocumentListener dl)
	{
		documentListeners.add(dl);
	}

	public void removeDocumentListener(DocumentListener dl)
	{
		documentListeners.remove(dl);
	}

	public void addUndoableEditListener(UndoableEditListener ul)
	{
		undoableEditListeners.add(ul);
	}

	public void removeUndoableEditListener(UndoableEditListener ul)
	{
		undoableEditListeners.remove(ul);
	}

	// PROPERTIES______________________________________________________________
	public Object getProperty(Object key)
	{
		return properties.get(key);
	}

	public void putProperty(Object key, Object value)
	{
		properties.put(key, value);
	}

	public Object removeProperty(Object key)
	{
		return properties.remove(key);
	}

	// TEXT_MANIPULATION_______________________________________________________
	public void remove(int start, int end)
	{
		XMLElement root = (XMLElement)super.getRootElement();
		Vector changed = root.getWrapper().remove(start,end);
	}

	/**
	  * Prepends the given string at the given offset in the text.
	  * The given attributes will be merged into the containing element
	  * attributes.
	  * @param <code>offset</code> the offset at which text will be prepended.
	  * @param <code>text</code> the text to be inserted.
	  * @param <code>attrs</code> the attributes that will be merged (but will
	  * never override the attributes already defined in the parent).
	  * <b>Important:</b> these attributes must be <code>XMLAttributes</code>
	  * to work.
	*/
	public void insertString(int offset,String text,AttributeSet attrs)
	{
		XMLElement root = (XMLElement)super.getRootElement();
		root.getWrapper().insert(text, offset, javax.swing.text.Position.Bias.Backward);
	}

	public String getText(int offset, int length)
	throws BadLocationException
	{
		return content.getString(offset, length);
	}
	
	/**
	  * Fetches the text contained within the given portion of the document. 
	  * If the partialReturn property on the txt parameter is false, the data
	  * returned in the Segment will be the entire length requested and may or
	  * may not be a copy depending upon how the data was stored. If the
	  * partialReturn property is true, only the amount of text that can be
	  * returned without creating a copy is returned. Using partial returns will
	  * give better performance for situations where large parts of the document
	  * are being scanned. 
	*/
	public void getText(int offset, int length, Segment s)
	throws BadLocationException
	{
		content.getChars(offset, length, s);
	}

	public int getLength()
	{
		XMLElement root = (XMLElement)super.getRootElement();
		if (root==null)
		{	return 0;	}
		XMLElementWrapper wrp = root.getWrapper();
		return wrp.getEndOffset()-wrp.getStartOffset();
	}

	// POSITIONING_____________________________________________________________
	public javax.swing.text.Position getStartPosition()
	{
		XMLElement root = (XMLElement)super.getRootElement();
		return new Position(root.getWrapper().getStartOffset());
	}

	public javax.swing.text.Position getEndPosition()
	{
		XMLElement root = (XMLElement)super.getRootElement();
		return new Position(root.getWrapper().getEndOffset());
	}

	public javax.swing.text.Position createPosition(int n)
	{
		Position p = new Position(n);
		positions.add(p);
		return p;
	}

	public javax.swing.text.Position createBackwardPosition(int n)
	{
		Position p = new Position(n, javax.swing.text.Position.Bias.Backward);
		positions.add(p);
		return p;
	}

	public javax.swing.text.Position createForwardPosition(int n)
	{
		Position p = new Position(n, javax.swing.text.Position.Bias.Forward);
		positions.add(p);
		return p;
	}
	// CONTENT_________________________________________________________________

	/**
	  * Returns the <code>Content</code> object that represents the text
	  * contained in this document.
	*/
	public GapContent getTextContent()
	{
		return content;
	}

	// MULTITHREAD_LOCK________________________________________________________
	public void render(Runnable renderer)
	{}

	protected  void writeLock()
	{}

	protected  void writeUnlock()
	{}

	//-------------------------------------------------------------------------
	//
	// Styled document
	//
	//-------------------------------------------------------------------------

	public Style addStyle(String nm, Style parent)
	{
		return null;
	}

	public Style getStyle(String nm) 
	{
		return null;
	}

	public void removeStyle(String nm) 
	{
	}

	public Element getCharacterElement(int pos) 
	{
		return getElementAt(pos);
	}

	public Color getBackground(AttributeSet attr) 
	{
		return new Color(200,200,200);
	}

	public Font getFont(AttributeSet attr) 
	{
		
		return font;
	}

	public Color getForeground(AttributeSet attr) 
	{
		return new Color(20,20,20);
	}

	public Style getLogicalStyle(int p) 
	{
		return null;
	}

	public void setCharacterAttributes(int offset, int length, AttributeSet s, boolean replace) 
	{
	}

	public void setLogicalStyle(int pos, Style s) 
	{
	}

	public void setParagraphAttributes(int offset, int length, AttributeSet s, boolean replace) 
	{
	}

	//-------------------------------------------------------------------------
	//
	// Tree model compliance
	//
	//-------------------------------------------------------------------------

	/**
	  * Returns the root of the tree.  Returns null only if the tree has
	  * no nodes.
	  * @return  the root of the tree
	*/
	public Object getRoot()
	{
		return ((XMLElement)super.getRootElement()).getWrapper();
	}

	/**
	  * Returns the child of <I>parent</I> at index <I>index</I> in the
	  * parent's child array.  <I>parent</I> must be a node previously obtained
	  * from this data source. This should not return null if <i>index</i> is a
	  * valid index for <i>parent</i> (that is <i>index</i> >= 0 &&
	  * <i>index</i> < getChildCount(<i>parent</i>)).  @param   parent  a node
	  * in the tree, obtained from this data source @return  the child of
	  * <I>parent</I> at index <I>index</I>
	*/
	public Object getChild(Object parent, int index)
	{
		if (parent instanceof XMLElement)
		{
			return ((XMLElement)parent).getWrapper().getElement(index);
		}
		else if (parent instanceof XMLElementWrapper)
		{
			return ((XMLElementWrapper)parent).getElement(index);
		}
		else
		{
			return null;
		}
	}

	/**
	  * Returns the number of children of <I>parent</I>.  Returns 0 if the node
	  * is a leaf or if it has no children.  <I>parent</I> must be a node
	  * previously obtained from this data source.
	  * @param   parent  a node in the tree, obtained from this data source
	  * @return  the number of children of the node <I>parent</I>
	*/
	public int getChildCount(Object parent)
	{
		if (parent instanceof XMLElement)
		{
			return ((XMLElement)parent).getWrapper().getElementCount();
		}
		else if (parent instanceof XMLElementWrapper)
		{
			return ((XMLElementWrapper)parent).getElementCount();
		}
		else
		{
			return 0;
		}
	}

	/**
	 * Returns true if <I>node</I> is a leaf.  It is possible for this method
	 * to return false even if <I>node</I> has no children.  A directory in a
	 * filesystem, for example, may contain no files; the node representing
	 * the directory is not a leaf, but it also has no children.
	 *
	 * @param   node    a node in the tree, obtained from this data source
	 * @return  true if <I>node</I> is a leaf
	 */
	public boolean isLeaf(Object node)
	{
		if (node instanceof XMLElement)
		{
			return ((XMLElement)node).getWrapper().isLeaf();
		}
		else if (node instanceof XMLElementWrapper)
		{
			return ((XMLElementWrapper)node).isLeaf();
		}
		else
		{
			return true;
		}
	}

	/**
	  * Messaged when the user has altered the value for the item identified
	  * by <I>path</I> to <I>newValue</I>.  If <I>newValue</I> signifies
	  * a truly new value the model should post a treeNodesChanged
	  * event.
	  * @param path path to the node that the user has altered.
	  * @param newValue the new value from the TreeCellEditor.
	*/
	public void valueForPathChanged(TreePath path, Object newValue)
	{
	}

	/**
	  * Returns the index of child in parent.
	*/
	public int getIndexOfChild(Object parent, Object child)
	{
		if (parent instanceof XMLElement)
		{
			return ((XMLElement)parent).getWrapper().getElementIndex(child);
		}
		else if (parent instanceof XMLElementWrapper)
		{
			return ((XMLElementWrapper)parent).getElementIndex(child);
		}
		else
		{
			return -1;
		}
	}

	/**
	  * Adds a listener for the TreeModelEvent posted after the tree changes.
	  * @see     #removeTreeModelListener
	  * @param   l       the listener to add
	*/
	public void addTreeModelListener(TreeModelListener l)
	{
		treeModelListeners.add(l);
	}

	/**
	  * Removes a listener previously added with <B>addTreeModelListener()</B>.
	  * @see     #addTreeModelListener
	  * @param   l       the listener to remove
	*/  
	public void removeTreeModelListener(TreeModelListener l)
	{
		treeModelListeners.remove(l);
	}

	//-------------------------------------------------------------------------
	//
	// Event management
	//
	//-------------------------------------------------------------------------

	protected void fireDocumentEvent(DocumentEvent evt)
	{
		Enumeration e=documentListeners.elements();
		while (e.hasMoreElements())
		{
			if ( evt.getType().equals(DocumentEvent.EventType.CHANGE) )
			{
				((DocumentListener)e.nextElement()).changedUpdate(evt);
			}
			else if ( evt.getType().equals(DocumentEvent.EventType.INSERT) )
			{
				((DocumentListener)e.nextElement()).insertUpdate(evt);
			}
			else if ( evt.getType().equals(DocumentEvent.EventType.REMOVE) )
			{
				DocumentListener lst =  (DocumentListener)e.nextElement();
				//System.out.println("Firing remove update to :"+lst);
				lst.removeUpdate(evt);
			}

		}
	}

	//-------------------------------------------------------------------------
	//
	// Tools
	//
	//-------------------------------------------------------------------------

	/**
	  * Returns a string rtpresenting the XML contained in this document. This
	  * string can be written to a file.
	*/
	public String toString()
	{
		XMLOutputter out = new XMLOutputter();
		return out.outputString(convertToJDOM());
	}

	/**
	  * Update all the position objects registered inside the document. This
	  * method is specific to the <code>XMLDocument</code>. It is called by the
	  * <code>fire</code> method of <code>DocumentEvent.ElementChange</code>
	  * objects.
	  * @param <code>offset</code> is the offset where the changed has occured
	  * 	   in the document. <code>offset >=0</code>
	  * @param <code>length</code> indicates the length of the change. If
	  * 	   <code>length>0</code> then the change is an insertion of
	  * 	   <code>length</code> characters from <code>offset</code>.<br/>In
	  * 	   case <code>length &lt;= 0</code> then the change is interpreted
	  * 	   as a deletion of <code>lenght</code> characters starting at
	  * 	   <code>offset</code>.
	*/
	public synchronized void updatePositions(int offset, int length)
	{
		ListIterator it = positions.listIterator();

		while (it.hasNext())
		{
			Position pos = (Position)it.next();
			//Case when text has been inserted
			if ( length>0 )
			{
				if ( (pos.offset==offset && pos.isForward())
				|| pos.offset>offset )
				{ pos.offset +=length; }
			}
			//Case when text has been removed
			else
			{
				if ( pos.offset>offset && pos.offset<offset-length )
				{	pos.offset = offset; }
				else if ( pos.offset>offset-length )
				{	pos.offset += length;	}
			}
		}
	}

	//------------------------------------------------------------------------
	//
	// Position
	//
	//------------------------------------------------------------------------

	/**
	  * Implements the Swing position interface in order to keep track of
	  * changes that occur in an XMLDocument. Basically a position will help
	  * keep track of a specific offset in the logical view (raw text) of the
	  * XMLDocument, taking into account future modification to the
	  * XMLDocument content.
	*/
	public static class Position implements javax.swing.text.Position
	{
		protected int offset;
		protected javax.swing.text.Position.Bias bias;

		public Position(int offset)
		{
			this.offset = offset;
			this.setBiasForward();
		}
		public Position(int offset, javax.swing.text.Position.Bias bias)
		{
			this.bias   = bias;
			this.offset = offset;
		}

		public int getOffset()
		{
			return offset;
		}

		public void setBiasForward()
		{
			this.bias = javax.swing.text.Position.Bias.Forward;
		}

		public void setBiasBackward()
		{
			this.bias = javax.swing.text.Position.Bias.Backward;
		}

		/**
		  * Tells if this position points on the character after the given
		  * offset.
		*/
		public boolean isForward()
		{
			return (this.bias == javax.swing.text.Position.Bias.Forward );
		}

		/**
		  * Tells if this position points on the character before the given
		  * offset.
		*/
		public boolean isBackward()
		{
			return (this.bias == javax.swing.text.Position.Bias.Backward );
		}

		public String toString()
		{
			StringBuffer result = new StringBuffer();
			result.append("Position offset:"+offset);
			if ( isForward() )
			{
				result.append(" points forwards.");
			}
			else
			{
				result.append(" points backwards.");
			}
			return result.toString()+" "+hashCode();
		} 
	}

	//------------------------------------------------------------------------
	//
	// DocumentEvent
	//
	//------------------------------------------------------------------------

	/**
	  * <p>The <code>DocumentEvent</code> class represents a set of modifications
	  * that have been done on the document. These modifications are located at
	  * a specific offset in the document and have a specific length.</p>
	  * <p>The modifications made can be either addition, removal or
	  * replacement of an element or element content. The aggregated
	  * <code>ElementEdit</code> contains information on which document nodes
	  * are affected by the document event.</p>
	*/
	public static class DocumentEvent implements javax.swing.event.DocumentEvent
	{
		// PUBLIC____________________________________________________________
		public static javax.swing.event.DocumentEvent.EventType INSERT =
		javax.swing.event.DocumentEvent.EventType.INSERT;

		public static javax.swing.event.DocumentEvent.EventType CHANGE =
		javax.swing.event.DocumentEvent.EventType.CHANGE;

		public static javax.swing.event.DocumentEvent.EventType REMOVE =
		javax.swing.event.DocumentEvent.EventType.REMOVE;


		// PRIVATE___________________________________________________________
		private int						offset;
		private int						length;
		private ElementEdit				edit;
		private DocumentEvent.EventType	type;
		private XMLDocument				doc;
		private String					value;

		// CONSTRUCTORS______________________________________________________

		/**
		  * Cretes a new document event that represents a removal of
		  * <code>len</code> characters starting at offset <code>offs</code> in
		  * document <code>doc.</code>.
		*/
		public DocumentEvent(XMLDocument doc, int offs, int len)
		{
			this.doc    = doc;
			this.offset = offs;
			this.length = len;
			this.type   = REMOVE;
			this.value  = null;
		}

		/**
		  * Cretes a new document event that represents an insertion of the
		  * <code>value</code> string at starting at offset <code>offs</code>
		  * in document <code>doc.</code>.
		*/
		public DocumentEvent(XMLDocument doc, int offs, String value)
		{
			this.doc    = doc;
			this.offset = offs;
			this.length = value.length();
			this.type   = INSERT;
			this.value  = value;
		}

		// METHODS___________________________________________________________

		/**
		  * This actually not adds an edit but <em>sets</em> the
		  * <code>ElementEdit</code>. Because a document event occurs at a
		  * specific location in the text only one element edit is really
		  * useful.
		  * @param edit Tells which element have been affected, and how, by
		  * this document event.
		*/
		public void addEdit(ElementEdit edit)
		{
			edit = edit;
		}

		public DocumentEvent.ElementChange getChange(Element elem) 
		{
			if (edit!=null && ( edit.getElement().equals(elem) ||
			edit.getElement().getParentElement().equals(elem)) )
			{	return edit; }
			else
			{	return null; }
		}

		public Document getDocument() 
		{
			return doc;
		}

		public int getLength() 
		{
			return length;
		}

		public int getOffset() 
		{
			return offset;
		}

		public DocumentEvent.EventType getType() 
		{
			return type;
		}

		public void fire()
		{
			if ( doc!=null )
			{
				doc.fireDocumentEvent(this);	
				if ( doc !=null && type==DocumentEvent.EventType.INSERT )
				{
					doc.updatePositions(offset, length);
					try
					{	doc.content.insertString(offset, value); }
					catch (BadLocationException e)
					{	e.printStackTrace(); }
					//System.out.println("Insertion of string:"+value+"\n at:"+offset);
				}
				//TODO: When it is a CHANGE
				else if( doc !=null && type==DocumentEvent.EventType.REMOVE )
				{
				
					
					doc.updatePositions(offset, -length);
					try
					{	
						String old = doc.content.getString(offset, length);
						doc.content.remove(offset, length);
					System.out.println("Removal of from:"+offset+" to:"+(offset+length));
					System.out.println(" was:"+old);	
					}
					catch (BadLocationException e)
					{	e.printStackTrace(); }
					
				}
				else
				{
					System.out.println("Element CHANGE not supported yet.");
				}
			}
		}

	}

	//------------------------------------------------------------------------
	//
	// DocumentEvent
	//
	//------------------------------------------------------------------------

	/**
	  * This class stores a single change made on an element of the document.
	  * There are three possibilities for the change:
	  * <ul>
	  *   <li>One or more children have been removed</li>
	  *   <li>One or more children have been added</li>
	  *   <li>One or more children have been changed</li>
	  * </ul>
	  * An element edit can be a combination of these three aspects.
	*/
	public static class ElementEdit implements DocumentEvent.ElementChange
	{
		// PRIVATE___________________________________________________________
		private Vector added;
		private Vector removed;
		private Element element;

		// CONSTRUCTORS______________________________________________________

		public ElementEdit(Element e)
		{
			element = e;
			added = new Vector();
			removed = new Vector();
		}

		// METHODS___________________________________________________________

		public void addAddedElement(Element elem)
		{
			added.add(elem);
		}

		public void addRemovedElement(Element elem)
		{
			removed.add(elem);
		}
		
		public Element[] getChildrenAdded() 
		{
			return (Element[])added.toArray();
		}

		public Element[] getChildrenRemoved() 
		{
			return (Element[])removed.toArray();
		}

		public Element getElement() 
		{
			return element;
		}

		/**
		  * Always returns 0.
		*/
		public int getIndex() 
		{
			return 0;
		}
	}
}

// EOF-Unix/ASCII-------------------------------------@RisingSun//Java//1.0//EN
