// ---------------------------------------------------@RisingSun//Java//1.0//EN
// Project			:	SPEdit-Prototypes
// ClassName		:	XMLElement
// Version			:	0.80
// Type				:	
//
// ThreadSafe		:	Nope
// Description		:	Subclass of JDOM Element, plus a Wrapper to make the
//						JDOM element visible as a Swing Element.
// URL				:	<http://SPEdit.sourceforge.net>
// ----------------------------------------------------------------------------
// Authors			:	Sebastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	11-Sep-2001	
// Last mod.		:	17-Oct-2001	
// History			:
//						17-Oct-2001 Added content monitoring methods to allow
//						the Wrapper to notify the containing document whenever
//						the element changes. Completed wrapping of JDOM
//						methods.(SP)
//						20-Sep-2001 Added methods for XMLText support.
//						14-Sep-2001 Implemented the text methods of the
//						wrapper- 'was hard ;). Implemented TreeNode
//						interface.(SP)
//						13-Sep-2001 Improved JavaDoc, plus added the
//						a lot of text related methods.(SP)
//						12-Sep-2001 Checking against test case.(SP)
//						11-Sep-2001 First implementation.(SP)
//
// Notes			:
//
// Bugs				:
//					  -
//
// To do			:
//					  - Add the setText method.
//					  - Make it threadsafe without synchronizing, like in every
//					  method that iterates. Make getStartOffset and
//					  getEndOffset synchronized.(SP)
//					  - Complete JavaDoc

package net.sourceforge.spedit.core;

import java.util.*;
import javax.swing.text.*;
import javax.swing.tree.TreeNode;
import org.jdom.CDATA;
import org.jdom.Text;
import org.jdom.Comment;
import org.jdom.EntityRef;
import org.jdom.ProcessingInstruction;

//----------------------------------------------------------------------------
//
// XMLElement
//
//----------------------------------------------------------------------------

/**
  * <p>The <code>XMLElement</code> is a subclass of the JDOM Element class that
  * allows it to wrapped by <code>XMLElementWrapper</code> objects.</p>
  * <p>The interest is that the XMLElement can be used both in a JDOM and in a
  * Swing document model, thus allowing the manipuation with the JDOM API to be
  * reflected visually by the Swing text representation mechanisms.</p>
  * <p>This class differs in some ways from its JDOM parent, it is a little bit
  * slower but notifies the containing <code>XMLDocument</code> of any ongoing
  * changes, thus allowing to use either the JDOM model or the Swing model for
  * document content manipulation.</p>
  * @version 0.80
  * @author  Sbastien Pierre
*/
public class XMLElement extends org.jdom.Element
{
	// PUBLIC_________________________________________________________________
	public static final int INITIAL_ARRAY_SIZE=5;

	// PROTECTED______________________________________________________________
	protected int            contentLength;
	protected int            startOffset;
	protected XMLElementWrapper wrapper;
	protected XMLAttributes     xmlAttributes;

	//-------------------------------------------------------------------------
	//
	// Constructors
	//
	//-------------------------------------------------------------------------

	/**
	  * Creates a new XMLElement with the name "default"
	*/
	public XMLElement()
	{
		this("default");
		init();
	}

	/**
	  * Creates a new XMLElement with the given name and parent.
	*/
	public XMLElement(String name)
	{
		super(name);
		init();
	}

	/**
	  * Creates a new XMLElement with the given name and parent.
	*/
	public XMLElement(String name, org.jdom.Namespace namespace) 
	{
		super(name,namespace);
		init();
	}

	/**
	  * Creates a new XMLElement with the given name and parent.
	*/
	public XMLElement(String name, String uri) 
	{
		super(name, uri);
		init();
	}

	/**
	  * Creates a new XMLElement with the given name and parent.
	*/
	public XMLElement (String name, String prefix, String uri) 
	{
		super(name, prefix, uri);
		init();
	}

	protected void init()
	{
		this.content = new ArrayList(5);
		this.wrapper = new XMLElementWrapper(this);
		xmlAttributes = new XMLAttributes(getAttributes());
	}

	//-------------------------------------------------------------------------
	//
	// Attribute monitoring
	//
	//-------------------------------------------------------------------------

	// The following methods are provided to be able to monitor changes in the
	// Element attributes. Chnaging attributes modifies the attributes list and
	// sometimes (if not always) recreates a new List, which render the
	// XMLAttributes obsolete. (SP)

	public org.jdom.Element addAttribute(org.jdom.Attribute attr)
	{
		org.jdom.Element elt = super.addAttribute(attr);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}

	public org.jdom.Element addAttribute(String name, String value)
	{
		org.jdom.Element elt = super.addAttribute(name, value);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}

	public org.jdom.Element setAttribute(org.jdom.Attribute attribute) 
	{
		org.jdom.Element elt = super.setAttribute(attribute);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}

	public org.jdom.Element setAttribute(java.lang.String name,
	java.lang.String value) 
	{
		org.jdom.Element elt = super.setAttribute(name, value);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}


	public org.jdom.Element setAttribute(java.lang.String name,
	java.lang.String value, org.jdom.Namespace ns) 
	{
		org.jdom.Element elt = super.setAttribute(name, value, ns);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}


	public org.jdom.Element setAttributes(java.util.List attributes) 
	{
		org.jdom.Element elt = super.setAttributes(attributes);
		xmlAttributes.setAttributes(getAttributes());
		return elt;
	}

	public boolean removeAttribute(java.lang.String name)
	{
		boolean result = super.removeAttribute(name);
		xmlAttributes.setAttributes(getAttributes());
		return result;
	}

	public boolean removeAttribute(java.lang.String name, org.jdom.Namespace
	ns) 
	{
		boolean result = super.removeAttribute(name,ns);
		xmlAttributes.setAttributes(getAttributes());
		return result;
	}


	public List getContent()
	{
		return content;
	}

	//-------------------------------------------------------------------------
	//
	// Content monitoring
	//
	//-------------------------------------------------------------------------

	public org.jdom.Element addContent(org.jdom.Element element) 
	{
		//We check that this is really an XMLElement
		if ( element instanceof XMLElement )
		{
			super.addContent(element);
			XMLElement el = (XMLElement)element;
			XMLDocument doc = (XMLDocument)getDocument();
			fireInsertEvent(el.getWrapper().getStartOffset(),
			el.getWrapper().getWholeText(), el.getWrapper());
		}
		//Otherwise we warn the user
		else
		{
			notSupported(element.getClass().getName());
		}
		return this;
	}

	public org.jdom.Element addContent(java.lang.String text) 
	{
		//We try to append the text to the last element, in case it is an
		//XMLText.
		int size = content.size();
		if (size > 0)
		{
			Object ob = content.get(size - 1);
			if (ob instanceof XMLText)
			{
				XMLText tobj = (XMLText)ob;
				//We get the offset at which the insertion will occur
				int offset = tobj.getWrapper().getEndOffset();
				//We do the insertion
				tobj.setValue(tobj.getValue()+text);
				//And notify the containing document
				fireInsertEvent(offset, text, tobj.getWrapper());
				return this;
			}
		}
		//Otherwise we create a new XMLText element and we add it.
		XMLText n = new XMLText(text);
		n.setParent(this);
		content.add(n);
		fireInsertEvent(n.getWrapper().getStartOffset(), text, n.getWrapper());
		return this;
	}

	public org.jdom.Element addContent(XMLText text)
	{
		//We try to append the text to an already existing the XMLText.
		int size = content.size();
		if (size > 0)
		{
			Object ob = content.get(size - 1);
			if (ob instanceof XMLText)
			{
				XMLText tobj = (XMLText)ob;
				//We get the offset at which the insertion will occur
				int offset = tobj.getWrapper().getEndOffset();
				//We append the text
				tobj.setValue(tobj.getValue()+text.getValue());
				//And notify the containing document
				fireInsertEvent(offset, text.getWrapper().getWholeText(),
				tobj.getWrapper());
				return this;
			}
		}
		//Otherwise we simply add it.
		content.add(text);
		text.setParent(this);
		fireInsertEvent(text.getWrapper().getStartOffset(),
		text.getWrapper().getWholeText(), text.getWrapper());
		return this;
	}

	/**
	  * Just the same as the addContent method, but this time no notice is made
	  * to the containing document. This is used by the <code>newContent</code>
	  * method.
	  * @param text The <code>XMLText</code> to be inserted.
	  * @see #newContent
	*/
	protected org.jdom.Element silentAddContent(XMLText text)
	{
		//We try to append the text to an already existing the XMLText.
		int size = content.size();
		if (size > 0)
		{
			Object ob = content.get(size - 1);
			if (ob instanceof Text)
			{
				XMLText tobj = (XMLText)ob;
				tobj.setValue(tobj.getValue()+text.getValue());
				return this;
			}
		}
		//Otherwise we simply add it.
		content.add(text);
		text.setParent(this);
		return this;
	}

	/**
	  * <p>This method is provided to set a whole new content that includes
	  * <code>XMLText</code> nodes in it. The current implementation (Beta7) of
	  * JDOM does not allow redefine the setContent method, so the actual
	  * is a kind of turnaround.</p>
	  * <p>Note that <em>this method does not notify the document</em> of
	  * changes that occured. This is important because every other addition or
	  * removal method notifies the containing document. This method allows the
	  * wrapper to modify the actual content without duplicating notifications,
	  * even allowing optimizations in the notification.</p>
	  * <p>If this seems unclear, just don't use this method
	  * <code>;)</code></p>
	  * @param content The new content of this <code>XMLElement</code>
	*/
	public void setNewContent(List content)
	{
		this.content = new ArrayList(INITIAL_ARRAY_SIZE);
		ListIterator it = content.listIterator();
		while( it.hasNext() )
		{
			Object o=it.next();
			if ( o instanceof org.jdom.Element )
			{	super.addContent((org.jdom.Element)o);	}
			else if ( o instanceof XMLText )
			{	silentAddContent((XMLText)o);	}
			else
			{
			System.err.println("Class "+o.getClass()+
			" not supported by XMLElement.setNewContent");
			}
		}
	}

	// UNSUPPORTED ELEMENTS____________________________________________________

	public org.jdom.Element addContent(Comment comment) 
	{	return notSupported("Comment"); }

	public org.jdom.Element addContent(EntityRef entity) 
	{	return notSupported("Entity reference"); }

	public org.jdom.Element addContent(ProcessingInstruction pi) 
	{	return notSupported("Processing instruction"); }

	public org.jdom.Element addContent(CDATA cdata) 
	{	return notSupported("CDATA"); }

	private org.jdom.Element notSupported(String type)
	{
		System.err.println("Child type "+type+
		" is not supported by XMLElement.");
		return this;
	}

	// TOOLS___________________________________________________________________

	/**
	  * Fires an insertion document event to the containing
	  * <code>XMLDocument</code> if it exists. This allows the
	  * <code>XMLElement</code> to notify the Swing document model of changes
	  * in the textual representation of the JDOM document.
	  * @param offset The offset at which the insertion occurs.
	  * @param text   The actual text that will be inserted.
	  * @param elt    The element concernced by the change.
	*/
	protected void fireInsertEvent(int offset, String text, Element elt)
	{
		XMLDocument.DocumentEvent event;
		XMLDocument doc = (XMLDocument)getDocument();
		//We fire a new document event in case the parent document exists
		if ( doc!=null )
		{
			event = new XMLDocument.DocumentEvent(doc, offset, text);
			event.addEdit(new XMLDocument.ElementEdit(elt));
			event.fire();
		}
	}

	//-------------------------------------------------------------------------
	//
	// Swing Element compliance
	//
	//-------------------------------------------------------------------------

	/**
	  * Returns the <b>Wrapper</b> attached to this <code>XMLElement</code>.
	  * This allow to make a bridge between the JDOM and the Swing Document
	  * Model.
	  * @return The associated Wrapper.
	*/
	public XMLElementWrapper getWrapper()
	{
		return wrapper;
	}

	/**
	  * Returns the document attached to this XMLElement.
	  * @return The belonging document.
	*/
	public org.jdom.Document getDocument()
	{
		return super.getDocument();
	}

	/**
	  * <p>Sets the parent document of this <code>XMLElement</code>. This does not
	  * automatically add the <code>XMLElement</code> to the document.
	  * @param  <code>doc</code> The belonging document which whould be an
	  * <code>XMLDocument</code> instance.</p>
	  * <p>This method has been redefined to make this method public rather
	  * than protected.</p>
	*/
	public void setDocument(XMLDocument doc)
	{
		super.setDocument(doc);
	}
}

// EOF-Unix/ASCII--------------------------------------@RisingSun//Java/1.0//EN
