// ---------------------------------------------------@RisingSun//Java//1.0//EN
// Project			:	SPEdit-Prototypes
// ClassName		:	XMLInformation
// Version			:	0.30
// Type				:	Core data class.
//						
// ThreadSafe		:	Nope.
// Description		:	Stores schema-level information on a specific element.
// URL				:	<http://SPEdit.sourceforge.net>
// ----------------------------------------------------------------------------
// Authors			:	Sbastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	19-Sep-2001	
// Last mod.		:	06-Nov-2001	
// History			:	
//						06-Nov-2001 Added attributes that indicate how the
//						element textual content should be processed.(sp)
//						01-Oct-2001 Started moving to the new stylesheetsheet
//						implementation.(SP)
//						26-Sep-2001 Implemented Style managment, has to switch
//						to Python for View instanciation.(SP)
//						24-Sep-2001 Implemented View instanciation, improved
//						JavaDoc. (SP)
//						19-Sep-2001 First implementation.(SP)
//
// Bugs				:
//						-
//
// To do			:
//						- Handle namespace in children
//						- getChildren uses a node to list insertable children.

package net.sourceforge.spedit.core;

import java.util.*;
import java.awt.Color;
import java.awt.Font;
import java.lang.reflect.*;
import org.jdom.*;
import javax.swing.text.View;

//-----------------------------------------------------------------------------
//
// XMLInformation
//
//-----------------------------------------------------------------------------

/**
  * <p>The <code>XMLInformation</code> class stores information related to a
  * specific element. This information is used by the application to know:
  * <ul>
  *   <li>Which view components can be used to represent this element inside
  *   the document.</li>
  *   <li>Which stylesheets are associated with this element, thus allowing to
  *   fine-tune the element appearance.</li>
  *   <li>What are the elements that can be added to a given element, in order
  *   to determine the possible insertions.</li>
  * </ul>
  * </p>
  * @author Sebastien Pierre
  * @version 0.30
*/
public class XMLInformation
{
	// PROTECTED_______________________________________________________________

	protected String		name;
	protected Namespace		namespace;
	protected Vector		children;
	/**
	  * The list of views supported by this element. The
	  * <code>defaultViewClass</code> is not parto of this list.
	*/
	protected Vector		attachedViewClasses;
	/**
	  * A hash that maps a set of attributes, expressed as an array of objects,
	  * to a the class of a <code>View</code> object.
	*/
	protected Hashtable		attachedViewParameters;
	protected Class			defaultViewClass;
	protected Styleable.StyleSheet	styleSheet;

	protected boolean		strip=true;	

	// CONSTRUCTORS____________________________________________________________

	/**
	  * Creates a new element information for the given element name.
	*/
	public XMLInformation( String name )
	{
		this.name = name;
		init();
	}

	/**
	  * This method is used by the constructors. It should not be called
	  * outside of the constructors.
	*/
	protected void init()
	{
		namespace = null;
		styleSheet = null;
		defaultViewClass = null;
		children = new Vector();
		attachedViewClasses = new Vector();
		attachedViewParameters = new Hashtable();
	}

	//------------------------------------------------------------------------
	//
	// Element identification
	//
	//------------------------------------------------------------------------

	/**
	  * Returns the name of the element.
	  * @return A <code>String</code> containing the element name.
	*/
	public String getName()
	{	return name;	}

	/**
	  * Returns the namespace of the element.
	  * @return The <code>Namespace</code> of the element, which can be set to
	  * <code>null</code>.
	*/
	public Namespace getNamespace()
	{	return namespace;	}

	//------------------------------------------------------------------------
	//
	// Stylesheet
	//
	//------------------------------------------------------------------------

	public Styleable.StyleSheet getStyleSheet()
	{	return styleSheet;	}
	
	//------------------------------------------------------------------------
	//
	// Element schema
	//
	//------------------------------------------------------------------------

	/**
	  * Tells wether or not the given element is allowed as a child of the
	  * current element. The element is supposed to be in the same namespace as
	  * the source element.
	  * @param <code>source</source> is the element in which the other element
	  * should be added.
	  * @param <code>name</code> is the name of the element to be added.
	  * @return <code>true</code> if the element is allowed, otherwise false.
	  * <code>null</code>.
	*/
	public boolean isElementAllowed(Element source, String name)
	{
		return true;
	}

	public void addChild(String element)
	{
		children.add(element);
	}

	/**
	  * Returns the possible children of this elements, as strings.
	*/
	public Enumeration getChildren()
	{
		return children.elements();
	}

	//------------------------------------------------------------------------
	//
	// View management
	//
	//------------------------------------------------------------------------

	/**
	  * Adds a view and its constructor attributes to this element information
	  * object.
	*/
	public void addView(Class view, Vector params)
	{
		if ( defaultViewClass==null )
		{
			defaultViewClass = view;
		}
		else
		{
			attachedViewClasses.add(view);
		}
		attachedViewParameters.put(view, params);
	}
	/**
	  * Instanciate a new <code>View</code> object that has been set as the
	  * default view for this kind of element.
	*/
	public View getNewDefaultView(javax.swing.text.Element elt)
	{
		if ( defaultViewClass==null )
		{
			System.err.println("No default view class for element : "
			+elt.getName());
		}
		return instanciateView(elt, defaultViewClass);
	}

	/**
	  * Instanciates the given <code>viewClass</code>. This class has to be
	  * part of the views associated to the element.
	  * @param <code>viewClass</code> the class of the view that has to be
	  * instanciated and configured. This class has to be either the
	  * <code>defaultViewClass</code> or to be part of the
	  * <code>attachedViewClasses</code>
	  * @return Returns <code>null</code> in case the <code>viewClass</code>
	  * was bad. Otherwise returns a new, properly configured, instance of the
	  * view class.
	*/
	protected View instanciateView(javax.swing.text.Element elt, Class viewClass)
	{
		//We check that the view class is declared in this Element information.
		if ( defaultViewClass==null || !(defaultViewClass.equals(viewClass) ||
		attachedViewClasses.contains(viewClass)) )
		{
		
			System.err.println("Unable to instanciate view class "+
			viewClass+" for element "+elt);
			System.err.println(toString());
			return null;
		}

		Object[]  obj = ((Vector)attachedViewParameters.get(viewClass)).toArray();
		Object[]  nobj = new Object[obj.length+1];
		Class[]   types = new Class[obj.length+1];

		nobj[0]   = elt;
		types[0] = javax.swing.text.Element.class;

		for (int i=0 ; i<obj.length ; i++)
		{
			nobj[i+1]  = obj[i];
			types[i+1] = obj[i].getClass();
		}

		try
		{
			Constructor cst = viewClass.getDeclaredConstructor(types);
			View instance = (View)cst.newInstance(nobj);
			if ( instance instanceof Styleable )
			{
				Styleable styleable = (Styleable)instance;
				String name=((XMLElementWrapper)nobj[0]).getName();
				Styleable.StyleSheet vs = new Styleable.StyleSheet(name, styleable);
				if ( styleSheet!=null )
				{	vs.inheritStyle(styleSheet);	}
				styleable.setStyleSheet(vs);
				//In case the styleable implementation does not update the
				//stylesheet we take care of it.
				vs.update();
			}
			return instance;
		}
		catch (Exception e)
		{
			System.err.println("Exception occured : "+e);
			e.printStackTrace();
			return null;
		}
	}
	
	//------------------------------------------------------------------------
	//
	// Textual content processing attributes
	//
	//------------------------------------------------------------------------

	public boolean hasStrip()
	{	return strip;	}
	
	//------------------------------------------------------------------------
	//
	// Tools
	//
	//------------------------------------------------------------------------

	public String toString()
	{
		StringBuffer result = new StringBuffer();
		result.append("XMLInformation for <"+namespace+":"+name+">\n");
		result.append(" Default view : "+defaultViewClass+"\n");
		Enumeration e;
		for ( e=attachedViewClasses.elements() ; e.hasMoreElements() ; )
		{
			result.append(" View         : "+e.nextElement()+"\n");
		}
		return result.toString();
	}


}

// EOF-Unix/ASCII-------------------------------------@RisingSun//Java//1.0//EN
