// Project			:	SPEdit-Prototypes
// ClassName		:	XMLViewFactory
// Version			:	0.80
// Type				:	Swing application component
//						
// ThreadSafe		:	Yep
// Description		:	Binds a Swing View to an XML element.
// URL				:	<http://SPEdit.sourceforge.net>
// ----------------------------------------------------------------------------
// Authors			:	Sbastien Pierre				<spierre@isb-sib.ch>
// ----------------------------------------------------------------------------
// Creation date	:	03-Sep-2001	
// Last mod.		:	07-Nov-2001	
// History			:	
//						07-Nov-2001 Addes strip attribute support.(sp)
//						18-Oct-2001 Added support for default stylesheet.
//						05-Oct-2001 Moved code from XMLInformation, updates to
//						keep in sync with the Configuration schema.
//						26-Sep-2001 Major rework of the configuration.(SP)
//						19-Sep-2001 Started implementing the XML configuration.
//						(SP)
//						03-Sep-2001 First implementation.(SP)
//
// Bugs				:
//						-
//
// To do			:
//						-Throw exceptions
//						-Handle namespaces

package net.sourceforge.spedit.core;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import javax.swing.*;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;
import org.jdom.*;
import org.jdom.input.*;
import net.sourceforge.spedit.ui.*;

//-----------------------------------------------------------------------------
//
// XMLViewFactory
//
//-----------------------------------------------------------------------------

/**
  * The <code>XMLViewFactory</code> associates a <code>View</code> to each
  * element of the document. These views can be configured using an XML
  * configuration file, as well as <code>XMLInformation</code> that are created
  * with each element.
  * @author Sbastien Pierre
  * @version 0.80
*/
public class XMLViewFactory implements ViewFactory
{
	// PROTECTED_______________________________________________________________
	protected Hashtable configs;
	protected Hashtable stylesheets;
	protected Hashtable views;
	protected Hashtable elementViews;

	//-------------------------------------------------------------------------
	//
	// Constructors
	//
	//-------------------------------------------------------------------------

	public XMLViewFactory()
	{
		configs = new Hashtable();
		stylesheets  = new Hashtable();
		views  = new Hashtable();
		elementViews  = new Hashtable();
	}

	//-------------------------------------------------------------------------
	//
	// Swing view methods
	//
	//-------------------------------------------------------------------------

	/**
	  * <p>This creates a <code>View</code> object bound to the given Swing
	  * element. This view is then wrapped around a <code>WrapperView</code>
	  * which adds CSS-like visual properties to the embedded views.</p>
	  * <p>When the new wrapper view is created it is registred as
	  * representing the given element. This allows to keep the association
	  * from elements to views.</p>
	  * @param  elt The element that the view will display.
	  * @return A new wrapped and registered view.
	  * @see #getElementView
	*/
	public View create(javax.swing.text.Element elt)
	{
		View view;

		if (elt==null)
		{	return null; }

		//We get the information and associate it to the element
		XMLInformation info = (XMLInformation)configs.get(elt.getName());
		((XMLNode)elt).setInformation(info);

		//We instanciate the view
		if ( configs.get(elt.getName())!=null )
		{	view = info.getNewDefaultView(elt); }
		else if ( elt instanceof XMLTextWrapper )
		{	view = new LabelView(elt); }
		else
		{	view = new BlockView(elt, BlockView.Y_AXIS); }

		//We create the new view and associate the stylesheet
		view = new WrapperView(view);
		if (info!=null && info.getStyleSheet()!=null)
		{
			Styleable.StyleSheet os = info.getStyleSheet();
			Styleable.StyleSheet ns = new Styleable.StyleSheet(
			os.getName(), (Styleable)view);
			ns.inheritStyle(os);
			((WrapperView)view).setStyleSheet(ns);
		}
		else
		{
			//We try to get a default stylesheet if it exists, which is
			//searched using the element name in normal and lowerase version
			Styleable.StyleSheet sheet = (Styleable.StyleSheet)stylesheets.get(elt.getName());
			if (sheet==null)
			{	stylesheets.get(elt.getName().toLowerCase()); }
			if (sheet==null)
			{
				((WrapperView)view).setStyleSheet(new
				Styleable.StyleSheet("unnamed",(Styleable)view));
			}
			else
			{
				((WrapperView)view).getStyleSheet().inheritStyle(sheet);
			}
		}

		//We register and return the view
		elementViews.put(elt, view);
		return view;
	}

	public View create(javax.swing.text.Element elt, String name)
	{
		Object[] vobj = (Object[])views.get(name);

		if ( vobj==null )
		{	return null;	}
		
		Class    vclass  = (Class)vobj[0];
		Vector   vparams = (Vector)vobj[1];

		Object[]  nobj = new Object[vparams.size()+1];
		Class[]   types = new Class[vparams.size()+1];

		nobj[0]   = elt;
		types[0] = javax.swing.text.Element.class;

		int i=0;
		for ( Enumeration e=vparams.elements() ; e.hasMoreElements() ; i++ )
		{
			Object o = e.nextElement();
			nobj[i+1]  = o;
			types[i+1] = o.getClass();
		}

		try
		{
			Constructor cst = vclass.getDeclaredConstructor(types);
			View instance = (View)cst.newInstance(nobj);
			return instance;
		}
		catch (Exception e)
		{
			System.err.println("Exception occured : "+e);
			e.printStackTrace();
			return null;
		}

	}

	public View getElementView(javax.swing.text.Element e)
	{
		return (View)elementViews.get(e);
	}

	public XMLInformation getElementInformation(javax.swing.text.Element e)
	{
		return (XMLInformation)configs.get(e.getName());
	}
	
	//------------------------------------------------------------------------
	//
	// XML Configuration methods
	//
	//------------------------------------------------------------------------

	/**
	  * Loads an XML configuration file and parses its content, merging it to
	  * the current XMLViewFactory.
	*/
	public void loadConfiguration(String filename)
	{
		try
		{
			SAXBuilder saxBuilder = XMLDocument.saxBuilder;
			File file = new File(filename);
			Document doc = saxBuilder.build(file);
			setConfiguration(doc.getRootElement());
		}
		catch(org.jdom.JDOMException e)
		{
			System.err.println("Unable to load the source XML file.");
			System.err.println("  "+e.toString());
			e.printStackTrace(System.err);
		}
	}
	/**
	  * Sets the configuration that contains the given element. This element
	  * must be named <code>Configuration</code>.
	*/
	public void setConfiguration(Element elt)
	{
		if (!elt.getName().equals("Configuration"))
		{
			System.err.println("Bad element given to XMLViewFactory.setConfiguration.");
			System.err.println(" Expected <Configuration>, found <"+
			elt.getName()+">");
			return;
		}

		//We do this for being threadsafe, because iteration over a list, which
		//is quicker, does not work at all in multithread.
		Styleable.StyleSheet stylesheet;
		for (int i=0 ; i<elt.getChildren().size() ; i++)
		{
			Element child = (Element)elt.getChildren().get(i);
			//We fill the hashtable with the different stylesheets present in the
			//Styles element.
			if ( child.getName().equals("Styles") )
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					stylesheet = createStyleSheet(gd_child);
					stylesheets.put(stylesheet.getName(), stylesheet);
				}
			}
		}

		//We create the views hash.
		for (int i=0 ; i<elt.getChildren().size() ; i++)
		{
			Element child = (Element)elt.getChildren().get(i);
			//We fill the hashtable with the different stylesheets present in the
			//Styles element.
			if ( child.getName().equals("Views") )
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					String name = gd_child.getAttribute("name").getValue();
					Object[] obj = createViewInformation(gd_child);
					views.put(name, obj);
				}
			}
		}

		//Then we create the information on the document content
		for (int i=0 ; i<elt.getChildren().size() ; i++)
		{
			Element child = (Element)elt.getChildren().get(i);
			//We fill the hashtable with the different stylesheets present in the
			//Styles element.
			if ( child.getName().equals("Document") )
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					XMLInformation info = createXMLInformation(gd_child);
					configs.put(info.getName(), info);
				}
			}
		}
	}

	//------------------------------------------------------------------------
	//
	// XML configuration
	//
	//------------------------------------------------------------------------

	/**
	  * Returns an array of two objects. The first one contains the reference
	  * to the view class described by the given JDOM <code>Element</code>, and
	  * the second the initialization attributes of this new view class.
	*/
	public Object[] createViewInformation(Element elt)
	{
		if (!elt.getName().equals("view"))
		{
			System.err.println("Bad element given to XMLInformation.updateElementInfo.");
			System.err.println(" Expected <view>, found <"+
			elt.getName()+">");
			return null;
		}

		Element child;
		Vector params = new Vector();

		//We create the view parameters list.
		for (int i=0 ; i<elt.getChildren().size() ; i++)
		{
			child = (Element)elt.getChildren().get(i);
			if (child.getName().equals("property"))
			{
				if (child.getAttribute("type").getValue().equals("string"))
				{
					params.add(new String(child.getAttributeValue("value")));
				}
				else if (child.getAttribute("type").getValue().equals("int"))
				{
					params.add(new Integer(Integer.parseInt(child.getAttributeValue("value"))));
				}
				else if (child.getAttribute("type").getValue().equals("float"))
				{
					params.add(new Float(Float.parseFloat(child.getAttributeValue("value"))));
				}
				else
				{
					System.err.println("Unknown property type in XMLInformation.updateElementInfo: "
					+child.getAttribute("type").getValue());
				}
			}
		}

		try
		{
			Class c = Class.forName(elt.getAttribute("class").getValue());
			Object[] res = new Object[2];
			res[0] = c;
			res[1] = params;
			return res;
		}
		catch (ClassNotFoundException e)
		{
			System.err.println("Unable to get view class "+elt.getAttribute("class").getValue());
			e.printStackTrace();
		}
		return null;
	}

	/**
	  *
	*/
	public XMLInformation createXMLInformation(Element elt)
	{
		if (!elt.getName().equals("element"))
		{
			System.err.println("Bad element given to XMLInformation.intialize.");
			System.err.println(" Expected <element>, found <"+
			elt.getName()+">");
			return null;
		}
		Element child;
		String name = elt.getAttribute("name").getValue();
		XMLInformation info = new XMLInformation(name);

		//We take care of the attributes
		if ( elt.getAttribute("strip")!=null )
		{
			String strip = elt.getAttribute("strip").getValue().toString().toLowerCase();
			if ( strip.equals("no") || strip.equals("yes") )
			{	info.strip = false;	}
			else
			{	info.strip = true;	}
		}
		//We take care of the children
		for (int i=0 ; i<elt.getChildren().size() ; i++)
		{
			child = (Element)elt.getChildren().get(i);
			//Here we set the view specific stylesheet object, which may be
			//inherited
			if (child.getName().equals("Style"))
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					if ( gd_child.getName().equals("use") )
					{
						String sname = gd_child.getAttribute("name").getValue();
						if ( stylesheets.get(sname)!=null )
						{
							Styleable.StyleSheet st = (Styleable.StyleSheet)stylesheets.get(sname);
							info.styleSheet = st;
						}
					}
				}
			}
			//We get the attached view definitions
			else if (child.getName().equals("Views"))
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					name = gd_child.getAttribute("name").getValue();
					Object[] vp = (Object[])views.get(name);
					if ( vp!=null )
					{
						info.addView((Class)vp[0], (Vector)vp[1]);
					}
				}
			}
			//We get the attached view definitions
			else if (child.getName().equals("Children"))
			{
				for (int j=0 ; j<child.getChildren().size() ; j++)
				{
					Element gd_child = (Element)child.getChildren().get(j);
					if ( gd_child.getName().equals("element") )
					{
						name = gd_child.getAttribute("name").getValue();
						info.addChild(name);
					}
				}
			}
		}
		return info;
	}

	//------------------------------------------------------------------------
	//
	// Style information
	//
	//------------------------------------------------------------------------

	public Styleable.StyleSheet createStyleSheet(Element elem)
	{
		if (!elem.getName().equals("Style"))
		{
			System.err.println("Bad element given to XMLInformation.initializeStyle.");
			System.err.println(" Expected <Style>, found <"+
			elem.getName()+">");
			return null;
		}
		
		ListIterator it = elem.getChildren().listIterator();
		Styleable.StyleSheet stylesheet = new Styleable.StyleSheet(elem.getAttribute("name").getValue());
		//We iterate through the stylesheet node children
		while ( it.hasNext() )
		{
			Element child = (Element)it.next();
			genStyle(stylesheet,"", child);
		}
		return stylesheet;
	}

	protected void genStyle(Styleable.StyleSheet stylesheet,String prefix, Element elem)
	{
		String nprefix;

		if ( prefix.equals("") )
		{ nprefix = elem.getName().toLowerCase(); }
		else
		{ nprefix = prefix+"."+elem.getName().toLowerCase(); }

		//Is it a group of properties ?
		if (elem.getAttributes().size()==0)
		{
			
			ListIterator it = elem.getChildren().listIterator();
			//We iterate through the stylesheet node children
			while ( it.hasNext() )
			{ genStyle(stylesheet,nprefix, (Element)it.next()); }
		}
		//Otherwise it is a property
		else
		{
			if ( elem.getAttribute("string")!=null )
			{
				stylesheet.setStyle(nprefix,elem.getAttribute("string").getValue());
			}
			else if ( elem.getAttribute("int")!=null )
			{
				Integer i = new Integer(elem.getAttribute("int").getValue());
				stylesheet.setStyle(nprefix,i);
			}
			else if ( elem.getAttribute("float")!=null )
			{
				Float i = new Float(elem.getAttribute("float").getValue());
				stylesheet.setStyle(nprefix,i);
			}
			else if ( elem.getAttribute("double")!=null )
			{
				Double i = new Double(elem.getAttribute("double").getValue());
				stylesheet.setStyle(nprefix,i);
			}
			else if ( elem.getAttribute("color")!=null )
			{
				stylesheet.setStyle(nprefix,Styleable.Utils.parseColor(elem.getAttribute("color").getValue()));
			}
		}
	}


}
// EOF-Unix/ASCII--------------------------------------@RisingSun//Java/1.0//EN
